#!/usr/bin/taintperl -w
#.........................................................................
# Copyright 1993, Jim Avera.  All Rights Reserved.
# 
# This work as a whole contains new material owned by the above copyright 
# holder, also contains portions derived from existing works used according
# to permission granted by the owners of those works.
#
# You are prohibited from copying, distributing, modifying, or using this
# work (or the portions owned by the above copyright holder) except as 
# described in the file "COPYRIGHT" which accompanies this program.
#..........................................................................

#
# slipcmd -- run or schedule commands which use a slip link
#
# USAGE:
#	slipcmd [-s /dev/slp] [-q] [-r] [-v] ifcname [cmd args...]
#
# Slipcmd runs or queues a command depending on whether a dial-up slip link 
# is currently established.
#
# -q causes the command to be deferred if the link is not established 
#    (i.e., no modem connection exists, even though IP believes the network
#    interface is up).  If the same command is attempted several times, only 
#    a single entry will be made in the deferred-work queue.
#
# -r causes any queued commands to be immediately executed, regardless of
#    link status.
#
# -v causes logging to stderr.
#
# Typically, work like polling a mail server is performed very frequently
# from cron using slipcmd -q, but occasionally using slipcmd -r (forcing 
# dial-out if necessary).
#
# To run deferred work as soon as the link is established for some other 
# reason, specify slipcmd -r as a "background command" in the parameter
# string given to slattach.  The slip dialing daemon (slipd) will then execute
# it whenever the link is established.
#
# EXIT VALUE:
#   0 if the command ran immediately (regardless of it's result).
#   1 if the command was queued (with -q) or discarded.
#   >1 if some error occurred.
# 
# NOTE: There should be an unused slip driver minor device so that this
#       program can find out if a dial-up link is currently established.
#	 For a single slip link, at least 3 devices must be configured 
#	 (for use by IP, slipd, and slipcmd).  If slipcmd can't open the
#	 slip driver it will assume the link is not active.
#
# Author: Jim Avera (jima@netcom.com)
#

require "getopts.pl";
require "shellwords.pl";
require "flush.pl";

$slipdpath = "@SBINDIR@/slipd";	#this is edited by Makefile

$ifcname = "sl0";
$qfiledir = "@SLIPCMD_SPOOLDIR@";

$verbose = 0;
$queue = 0;
$forcerun = 0;

$progname = "slipcmd";	# suid scripts don't know their name

$opt_v = $opt_q = $opt_r = 0;
$opt_s = "/dev/slp";
&Getopts('s:vqr') || &usage();
$verbose = 1 	   if $opt_v;
$queue = 1 	   if $opt_q;
$forcerun = 1 	   if $opt_r;
$slippath = $opt_s;

umask(0022);

$ENV{'PATH'} = '/usr/bin:/bin';
$ENV{'IFS'} = '';
$ENV{'SHELL'} = '/bin/sh';

if (@ARGV < 1) { &usage(); }
&untaint( $ifcname = shift(@ARGV) );

$n = system( "$slipdpath -s $slippath -c $ifcname" );
if    ($n==0) 	     { $inuse = 1; }
elsif ($n==(1 << 8)) { $inuse = 0; }
else { die "$progname: $slipdpath failed ($n), exiting\n"; }
	
if (@ARGV) {
    # Form a single string for the queue file, containing:
    #   user/uid group/gid commandpath quoted-arg0 quoted-arg1 ...
    (($username)  = getpwuid($<)) || ($username = $<);
    (($groupname) = getgrgid($()) || ($groupname = $();
    $cmd = "$username $groupname $ARGV[0]";
    $ARGV[0] =~ s@.*/@@;
    while ($_ = shift @ARGV) {
        if (/\W/ && !/^\s*(['"]).*\1$/) {
    	$cmd .= " '$_'";
        } else {
    	$cmd .= " $_";
        }
    }

    if ($forcerun) {
    	# do it as though part of the queue we are about to run
        &runcmd($cmd);
    }
    elsif ($inuse) {
    	# The link is active - run it now
        &runcmd($cmd);
    } else {
        if ($queue) {
    	    # Put command on the deferred-work queue
    	    &queuecmd($ifcname, $cmd);
        } else {
    	    # Forget it completely
    	    print STDERR "$progname: $ifcname not active - cmd was not run\n"
    	        unless !$verbose;
        }
    }
}

if ($forcerun) { &runqueue($ifcname); }

exit(0) if $inuse;
exit 1;
# --- END OF MAIN PROGRAM ---


# Put a command string on the deferred-work queue, if not there already.
sub 
queuecmd {
	local($ifcn, $cmd) = @_;
	local($_, $qfilepath, *QFILE);
		
	&untaint( $ifcn );
	$qfilepath = "$qfiledir/$ifcn";

	# Scan the existing file to see if the cmd is already there.
	# If not, append the line (assumes atomic append).  

        if (open(QFILE,$qfilepath)) {
	    while (<QFILE>) {
		chop;
		if ($_ eq $cmd) {
		    print STDERR "$progname: cmd already on queue ($cmd)\n"
			unless (!$verbose);
		    close(QFILE);
		    return;
		}
	    }
	}
	mkdir($qfiledir, 0755);
	&getlock("$qfilepath.L");
        open(QFILE,">>$qfilepath") || die "Can't create $qfilepath:$!";
	print(QFILE $cmd, "\n");
	close(QFILE);
	&freelock("$qfilepath.L");

	print STDERR "$progname: cmd deferred ($cmd)\n" unless (!$verbose);
}

# Run the stored queue
sub runqueue {
	local($ifcname) = @_;
	local($_,$qfilepath,*QFILE);
		
	$qfilepath = "$qfiledir/$ifcname";

	# Read the command strings into an associative array, thereby
	# eliminating any redundancies (created by processes which
	# did not use the proper locking protocol).

	&getlock("$qfilepath.L");
        if (open(QFILE,$qfilepath)) {
	    unlink $qfilepath;
	    &freelock("$qfilepath.L");
	    while (<QFILE>) {
		&runcmd($_);
	    }
	    close(QFILE);
	} else {
	    &freelock("$qfilepath.L");
	}
}

# Execute a command.  The input string is encoded as
#   user/uid group/gid progpath av0 cmdstring...
sub 
runcmd {
	local($cmd) = @_;
	local(@F,$uid,$gid,$path,$pid);
	@F = &shellwords($cmd);
	if (@F < 4) {
	    printf STDERR "$progname: CORRUPT QUEUE RECORD:$cmd\n";
	    return 0;
	}
	$uid = shift @F;
	if ($uid !~ /^\d+$/) {
	    ($d,$d,$uid) = getpwnam($uid);
	    (defined($uid)) || die "Unknown user \"$uid\"\n";
	}
	$gid = shift @F;
	if ($gid !~ /^\d+$/) {
	    ($d,$d,$gid) = getgrnam($gid);
	    (defined($gid)) || die "Unknown user \"$gid\"\n";
	}
	$path = shift @F;
	print STDERR "$progname: +[$uid,$gid] $path @F\n" unless !$verbose;
	if ($pid = fork) {
	    # parent
	    (wait == $pid) || die("unexpected pid from wait\n");
	} else {
	    # child -- become the user and run the program
	    $) = $gid;
	    $> = $uid;
	    exec $path $F[0],@F[1..$#F];
	    die "Cant exec $path:$!\n";
	}
}

sub usage {
	print STDERR "USAGE:\n";
	print STDERR "  $progname -q ifcname cmd args...   #runs cmd or queues it if link not active\n";
	print STDERR "  $progname -r ifcname [cmd args...] #unconditionally executes the queue\n";
	print STDERR "  Other options: -v\n";
	print STDERR "                 -s $slippath\n";
	exit(2);
}

sub
untaint {
	$_[0] =~ m/(.*)/;
	$_[0] = $1;
	return $1;
}

sub
getlock {
	local($path) = @_;
	local(*F, $lockpid, $n);
	# Perl doesn't support O_EXCL so we can't create a lock file only
	# if it doesn't already exist.  Therefore we create a directory
	# as a meta-lockfile, just to protect us while we write our pid
	# into the regular lock file.
	TRY: for (;;) {
	    &_meta_lock($path);
	    if (open(F, "+<$path")) {
		# File exists -- check to see if pid is valid
	        seek(F,0,0);
	        $lockpid = <F>;
	        if ($lockpid =~ /^\s*\d+\s*$/) {
		    if (kill(0,$lockpid) || ($! =~ /owner|perm/)) {
		        # sent nop signal or got permission error
		        &_meta_unlock($path);
		        print STDERR "BUSY(pid $lockpid): $path\n"
			    unless !$verbose;
		        sleep 1;
		        next TRY;
		    } else {
		        print STDERR "STALE LOCK(pid $lockpid): $path\n"
			    unless !$verbose;
		    }
	        }
	        seek(F,0,0);
	    } else {
	        open(F, ">$path") 
	          || (&_meta_unlock($path), die "Can't create $path:$!\n");
	    }
	    printf F $$;
	    close(F);
	    &_meta_unlock($path);
	    print STDERR "$progname: LOCKED $path\n" unless !$verbose;
	    last;
	}
}

sub
freelock {
	local($path) = @_;
	local($n);
	&_meta_lock($path);
	$n = unlink($path);
	&_meta_unlock($path);
	(!$n) && die "Can't unlink $path:$!\n";
	print STDERR "$progname: UNLOCKED $path\n" unless !$verbose;
}

sub _meta_lock_sighand {
	$got_sig = $_[0];
	print STDERR "Got SIG$got_sig\n" unless !$verbose;
}
sub _meta_lock {
	local($lockpath) = @_;
	local($lockdir) = $lockpath . ".d";
	local($_,%sigs);

	$got_sig = 0;
	foreach $_ ('HUP', 'INT', 'QUIT', 'TERM', 'TSTP') {
	    (!defined $SIG{$_}) && next;
	    $sigs{$_} = $SIG{$_};
	    $SIG{$_} = '_meta_lock_sighand';
	}
	while (!mkdir($lockdir,0600)) {
	    ($! !~ m/exist/) && die "$lockdir:$!\n";
	    sleep 1;
	    if ($got_sig) {
	        foreach $_ (keys %sigs) { $SIG{$_} = $sigs{$_}; }
	        kill($$, $got_sig);
	    }
	}
	foreach $_ (keys %sigs) { $SIG{$_} = $sigs{$_}; }
}
sub _meta_unlock {
	local($lockpath) = @_;
	local($lockdir) = $lockpath . ".d";
	rmdir($lockdir) || die "rmdir $lockdir:$!\n";
}
